﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using CQG.Operators;
using CQG.Operators.Primitives;


namespace SampleMPVA
{
    /// <summary>
    /// Operator for Sample Market Profile Area Study (MPVA) study.
    /// <para>
    ///     Expression inputs:
    ///     <list type="bullet">
    ///         <item>Bar Low value</item>
    ///         <item>Bar High value</item>
    ///         <item>Index of bar from the day start</item>
    ///         <item>Index of bar from the session start</item>
    ///         <item>Indicates whether the calculated bar is current bar</item>
    ///         <item>Bar time</item>
    ///     </list>
    /// </para>
    /// <para>
    ///     Expression parameters:
    ///     <list type="bullet">
    ///         <item>color0 is a color of session # 0</item>
    ///         <item>color1 is a color of session # 1</item>
    ///         <item>color2 is a color of session # 2</item>
    ///         <item>pocColor is a color of POC line</item>
    ///     </list>
    ///  </para>
    /// </summary>
    [CQGOperator("SampleMPVA(@, @, @, @, @, @, color0, color1, color2, pocColor)", FillInvalids=false)]
    public class SampleMPVA : IOperator
    {
        /// <summary>
        /// Operator state
        /// </summary>
        private class State : IDeepCloneable
        {
            /// <summary>
            /// Keeps day index of current bar.
            /// We know that new bar becomes current when this index is changed.
            /// </summary>
            public int CurrentBarIndex;

            /// <summary>
            /// Keeps time of current bar. This time is used to rebuild chart when the current bar is changed.
            /// </summary>
            public DateTime CurrentBarTime;

            public object DeepClone()
            {
                return MemberwiseClone();
            }
        }

        /// <summary>
        /// Keeps information about profile
        /// </summary>
        private class Profile
        {
            /// <summary>
            /// Index in inputs list where the profile starts
            /// </summary>
            public readonly int StartIndex;

            /// <summary>
            /// Minimum profile price
            /// </summary>
            public readonly double MinPrice;

            /// <summary>
            /// Count of sessions the profile contains
            /// </summary>
            public readonly int SessionsCount;

            /// <summary>
            /// Collection of TPO counts for each value and each session. The first dimension is price index 
            /// (starting from MinPrice with _tickSize step). The second dimension is session index (it has 
            /// size = SessionsCount + 1 - the last element contains sum of TPO count for all sessions).
            /// </summary>
            public readonly int[,] Values;

            /// <summary>
            /// POC price
            /// </summary>
            public double PoC = double.NaN;

            public Profile(int startIndex, double minPrice, int valuesCount, int sessionsCount)
            {
                StartIndex = startIndex;
                MinPrice = minPrice;
                SessionsCount = sessionsCount;
                Values = new int[valuesCount, sessionsCount + 1];
            }
        }

        /// <summary>
        /// Extra bars count we need to calculate the profile. Such big number allows to calculate
        /// profiles for minute bars and non-timed chart types like CVB.
        /// </summary>
        private int _stepCount = 2000;
        
        /// <summary>
        /// Tick size
        /// </summary>
        private double _tickSize = 1.0;

        /// <summary>
        /// The constant is used for converting time received from operator input
        /// </summary>
        private static DateTime _year1900 = new DateTime(1900, 1, 1);

        /// <summary>
        /// Session colors
        /// </summary>
        private List<Color> _colors = new List<Color>
        {
            Color.Olive,
            Color.Blue,
            Color.Brown,
        };

        /// <summary>
        /// POC Color
        /// </summary>
        private Color _pocColor = Color.Red;

        /// <summary>
        /// Reference to <see cref="IBarsRebuilder"/> object.
        /// </summary>
        private IBarsRebuilder _barsRebuilder;


        #region IOperator implementation

        /// <summary>
        /// Implements <see cref="IOperator.Initialize"/>
        /// </summary>
        public int Initialize(IDictionary<string, string> i_params, IBarsRebuilder i_barsRebuilder)
        {
            if (!double.TryParse(i_params["-MinimumValueChange"], NumberStyles.Float, CultureInfo.InvariantCulture, out _tickSize) ||
                _tickSize <= 0)
            {
                throw new ArgumentException("-MinimumValueChange value is not valid");
            }

            uint colorref;
            if (uint.TryParse(i_params["color0"], out colorref))
            {
                _colors[0] = colorFromCOLORREF(colorref);
            }

            if (uint.TryParse(i_params["color1"], out colorref))
            {
                _colors[1] = colorFromCOLORREF(colorref);
            }

            if (uint.TryParse(i_params["color2"], out colorref))
            {
                _colors[2] = colorFromCOLORREF(colorref);
            }

            if (uint.TryParse(i_params["pocColor"], out colorref))
            {
                _pocColor = colorFromCOLORREF(colorref);
            }

            _barsRebuilder = i_barsRebuilder;

            return _stepCount;
        }

        /// <summary>
        /// Implements <see cref="IOperator.Calculate"/>
        /// </summary>
        public void Calculate(IList<double>[] i_inputs, double[] io_outputs, ref IDeepCloneable io_state, IList<object> io_primitives)
        {
            bool isCurrentBar = i_inputs[4][_stepCount - 1] == 1.0;

            if (isCurrentBar || i_inputs[2][_stepCount - 1] == 0.0)
            {
                // Profile for the current day we place on the current bar. When another bar becomes current
                // we should first remove all primitives added on current bar last time and then add them on new current bar.
                // To detect such situation we use calculation state.

                if (isCurrentBar)
                {
                    State state = (State)io_state;
                    if (state != null)
                    {
                        // If new current bar is appeared call RebuildRange from the previous current bar 
                        if (state.CurrentBarIndex != i_inputs[2][_stepCount - 1])
                        {
                            _barsRebuilder.RebuildRange(state.CurrentBarTime, DateTime.MaxValue);

                            state.CurrentBarIndex = (int)i_inputs[2][_stepCount - 1];
                            state.CurrentBarTime = _year1900 + TimeSpan.FromMinutes(i_inputs[5][_stepCount - 1]);

                            return;
                        }
                    }
                    else
                    {
                        io_state = new State();
                        state = (State)io_state;

                        state.CurrentBarIndex = (int)i_inputs[2][_stepCount - 1];
                        state.CurrentBarTime = _year1900 + TimeSpan.FromMinutes(i_inputs[5][_stepCount - 1]);
                    }
                }

                Profile profile = calculateProfile(i_inputs[0], i_inputs[1], i_inputs[2], i_inputs[3], isCurrentBar);

                // Add primitives to visualize the profile
                double halfTickSize = _tickSize * 0.5;
                int xBase = profile.StartIndex - _stepCount + 1;

                int valuesCount = profile.Values.Length / (profile.SessionsCount + 1);
                for (int i = 0; i < valuesCount; ++i)
                {
                    int x = xBase;
                    double price = profile.MinPrice + i * _tickSize;

                    for (int n = 0; n < profile.SessionsCount; ++n)
                    {
                        int colorIndex = n % _colors.Count;
                        int value = profile.Values[i, n];
                        if (value > 0)
                        {
                            BorderPrimitive prim = new BorderPrimitive();
                            prim.BackgroundColor = _colors[colorIndex];

                            prim.Point1 = new ChartPosition(new ChartTime(x), price + halfTickSize);
                            prim.Point2 = new ChartPosition(new ChartTime(x + value), price - halfTickSize);

                            prim.Tooltip = String.Format("Price = {0}\nSession # = {1}\nDay TPO Count = {2}\nSession TPO Count = {3}",
                                price.ToString(CultureInfo.InvariantCulture), n,
                                profile.Values[i, profile.SessionsCount], profile.Values[i, n]);

                            io_primitives.Add(prim);

                            x += value;
                        }
                    }
                }

                // Add line for Point of Control
                if (!Double.IsNaN(profile.PoC))
                {
                    LinePrimitive line = new LinePrimitive();
                    line.EndsInPoint1 = true;
                    line.EndsInPoint2 = !isCurrentBar;
                    line.Point1 = new ChartPosition(new ChartTime(xBase), profile.PoC);
                    line.Point2 = new ChartPosition(new ChartTime(), profile.PoC);
                    line.Pen = new PrimitivePen(_pocColor, 1, 0.5); // PS_DASH

                    line.Tooltip = String.Format("POC = {0}", profile.PoC.ToString(CultureInfo.InvariantCulture));

                    io_primitives.Add(line);
                }
            }
        }

        #endregion

        /// <summary>
        /// Calculate profile
        /// </summary>
        private Profile calculateProfile(IList<double> minValues, IList<double> maxValues,
            IList<double> barsFromDayStart, IList<double> barsFromSessionStart, bool isCurrentDay)
        {
            double minPrice = double.MaxValue;
            double maxPrice = double.MinValue;

            // For historical profiles we place primitives on the first bar of the next day, so the current bar 
            // shouldn't be included in profile calculation. But for current day profile we place primitives on 
            // the last bar that should be included in profile calculation.
            int startIndex = barsFromSessionStart.Count - (isCurrentDay ? 1 : 2);
            int sessionsCount = 1;

            // 1. Calculate profile range
            for (int i = startIndex; i >= 0; --i)
            {
                if (!Double.IsNaN(minValues[i]) && !Double.IsNaN(maxValues[i]))
                {
                    minPrice = Math.Min(minPrice, minValues[i]);
                    maxPrice = Math.Max(maxPrice, maxValues[i]);
                }

                if (barsFromDayStart[i] == 0.0)
                {
                    startIndex = i;
                    break;
                }

                if (barsFromSessionStart[i] == 0.0)
                {
                    ++sessionsCount;
                }
            }

            int valuesCount = (int)Math.Round((maxPrice - minPrice) / _tickSize) + 1;

            Profile profile = new Profile(startIndex, minPrice, valuesCount, sessionsCount);

            // 2. Fill profile values
            int sessionIndex = sessionsCount - 1;
            for (int i = barsFromSessionStart.Count - (isCurrentDay ? 1 : 2); i >= 0; --i)
            {
                if (!Double.IsNaN(minValues[i]) && !Double.IsNaN(maxValues[i]))
                {
                    int lowValueIndex = (int)Math.Round((minValues[i] - minPrice) / _tickSize);
                    int highValueIndex = (int)Math.Round((maxValues[i] - minPrice) / _tickSize);

                    for (int valueIndex = lowValueIndex; valueIndex <= highValueIndex; ++valueIndex)
                    {
                        ++profile.Values[valueIndex, sessionIndex];
                    }
                }

                if (barsFromDayStart[i] == 0.0)
                {
                    break;
                }

                if (barsFromSessionStart[i] == 0.0)
                {
                    --sessionIndex;
                }
            }

            // 3. Calculate Point of Control
            int pocCount = -1;
            int pocIndex = -1;
            int medianIndex = valuesCount / 2;

            for (int i = 0; i < valuesCount; ++i)
            {
                int count = 0;
                for (int j = 0; j < sessionsCount; ++j)
                {
                    count += profile.Values[i, j];
                }

                profile.Values[i, sessionsCount] = count;

                if (count > pocCount ||
                    count == pocCount && Math.Abs(i - medianIndex) < Math.Abs(pocIndex - medianIndex))
                {
                    pocCount = count;
                    pocIndex = i;
                }
            }

            if (pocIndex >= 0)
            {
                profile.PoC = profile.MinPrice + pocIndex * _tickSize;
            }

            return profile;
        }

        /// <summary>
        /// Converts COLORREF value (passed as color parameter from CQGIC) to <see cref="Color"/> value.
        /// </summary>
        private static Color colorFromCOLORREF(uint color)
        {
            uint alpha = 0xFF000000 - (color & 0xFF000000);
            uint red = (color & 0xFF) << 16;
            uint green = (color & 0xFF00);
            uint blue = (color & 0xFF0000) >> 16;

            return Color.FromArgb((int)(alpha | red | green | blue));
        }

    }
}
