﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using CQG;

namespace SampleOrderDisplay
{
    /// <summary>
    /// Defines order subscriber - the consumer of <see cref="CELWrapper"/> class.
    /// </summary>
    internal interface IOrdersSubscriber
    {
        /// <summary>
        /// Called by <see cref="CELWrapper"/> when orders related to the subscriber's instrument are changed.
        /// </summary>
        /// <param name="ordersList">Updated list of all known orders for the subscribers' instrument</param>
        /// <param name="startTime">
        /// This value is set when only one order is changed. The value specifies start time of order period.
        /// </param>
        void UpdateOrders(IList<OrderInfo> ordersList, DateTime? startTime = null);

        /// <summary>
        /// Called by <see cref="CELWrapper"/> if some error occures.
        /// </summary>
        /// <param name="errorMessage">The error message</param>
        void OnError(string errorMessage);
    }

    /// <summary>
    /// Singleton wrapper over <see cref="CQG.CEL"/> to be used by <see cref="SampleOrderDisplay"/> operators.
    /// </summary>
    /// <remarks>
    /// Since <see cref="CQG.CEL"/> has a limitation for a number of instances created, we can't allow each
    /// operator instance to own a separate instance of <see cref="CQG.CEL"/>. So it's shared by all operators
    /// in a single singleton class.
    /// </remarks>
    internal class CELWrapper
    {
        /// <summary>
        /// Contains data related to one instrument
        /// </summary>
        private class InstrumentData
        {
            /// <summary>
            /// Instrument Id
            /// </summary>
            public int InstrumentId;

            /// <summary>
            /// List of order subscribers for the given instrument
            /// </summary>
            public List<IOrdersSubscriber> Subscribers;

            /// <summary>
            /// Collection of order information received for the given instrument
            /// </summary>
            public List<OrderInfo> Orders;

            /// <summary>
            /// 
            /// </summary>
            public List<CQGOrdersQuery> Queries;

            /// <summary>
            /// Initializes a new instance of <see cref="InstrumentData"/> object for the specified instrument.
            /// </summary>
            public InstrumentData(int instrumentId)
            {
                InstrumentId = instrumentId;
                Subscribers = new List<IOrdersSubscriber>();
                Orders = new List<OrderInfo>();
                Queries = new List<CQGOrdersQuery>();
            }

            /// <summary>
            /// Calls <see cref="IOrdersSubscriber.UpdateOrders"/> on all related order subscribers.
            /// </summary>
            /// <param name="startTime">See startTime parameter in <see cref="IOrdersSubscriber.UpdateOrders"/> method.</param>
            public void UpdateOrders(DateTime? startTime = null)
            {
                foreach (IOrdersSubscriber subscriber in Subscribers)
                {
                    subscriber.UpdateOrders(Orders, startTime);
                }
            }
        }

        /// <summary>
        /// Singleton instance of this class.
        /// </summary>
        private static CELWrapper _instance;

        /// <summary>
        /// The <see cref="CQGCEL"/> object used to access CQG API.
        /// </summary>
        private CQGCEL _cel;
        
        /// <summary>
        /// Contains instrument data for each subscribed instrument.
        /// </summary>
        private List<InstrumentData> _instrumentData = new List<InstrumentData>();

        /// <summary>
        /// For each symbol that is not yet resolved contains list of subscribers for this symbol
        /// </summary>
        private Dictionary<string, List<IOrdersSubscriber>> _waitingSubscriptions = new Dictionary<string, List<IOrdersSubscriber>>();

        /// <summary>
        /// Hides the constructor, don't allow to create <see cref="CELWrapper"/> instances externally.
        /// </summary>
        private CELWrapper()
        {}

        /// <summary>
        /// Gets instance of <see cref="CELWrapper"/> singleton.
        /// </summary>
        public static CELWrapper GetInstance()
        {
            if (_instance == null)
            {
                _instance = new CELWrapper();
            }
            return _instance;
        }

        /// <summary>
        /// Adds order subscriber
        /// </summary>
        /// <param name="subscriber">The order subscriber to add</param>
        /// <param name="symbol">Instrument symbol. Only orders related to this instrument will be subscribed.</param>
        public void SubscribeOrders(IOrdersSubscriber subscriber, string symbol)
        {
            if (_cel == null)
            {
                initializeCel();
            }

            List<IOrdersSubscriber> subscribers;
            if (!_waitingSubscriptions.TryGetValue(symbol, out subscribers))
            {
                subscribers = new List<IOrdersSubscriber>();
                _waitingSubscriptions.Add(symbol, subscribers);

                if (_cel.IsStarted)
                {
                    requestInstrument(symbol);
                }
            }

            subscribers.Add(subscriber);
        }

        /// <summary>
        /// Removes order subscriber added earlies with <see cref="SubscribeOrders"/> method.
        /// </summary>
        /// <param name="subscriber">The order subscriber to unsubscribe</param>
        public void UnsubscribeOrders(IOrdersSubscriber subscriber)
        {
            foreach (var pair in _waitingSubscriptions)
            {
                if (pair.Value.Remove(subscriber))
                {
                    if (pair.Value.Count == 0)
                    {
                        _waitingSubscriptions.Remove(pair.Key);
                    }
                    break;
                }
            }

            foreach (var data in _instrumentData)
            {
                if (data.Subscribers.Remove(subscriber))
                {
                    // Remove the instrument data if there is no subscriber left
                    if (data.Subscribers.Count == 0)
                    {
                        foreach (CQGOrdersQuery query in data.Queries)
                        {
                            query.Cancel();
                        }

                        CQGInstrument instrument = _cel.Instruments[data.InstrumentId];
                        if (instrument != null)
                        {
                            _cel.RemoveInstrument(instrument);
                        }

                        _instrumentData.Remove(data);
                    }
                    break;
                }
            }

            // Shutdown CQGCEL if there is no subscribers anymore
            if (_waitingSubscriptions.Count == 0 && _instrumentData.Count == 0)
            {
                _cel.Shutdown();
                _cel = null;
            }
        }

        /// <summary>
        /// Initializes CQGCEL object.
        /// </summary>
        private void initializeCel()
        {
            _cel = new CQGCEL();

            _cel.APIConfiguration.ReadyStatusCheck = eReadyStatusCheck.rscOff;
            _cel.APIConfiguration.CollectionsThrowException = false;
            _cel.APIConfiguration.NewInstrumentMode = true;
            _cel.APIConfiguration.DefPositionSubscriptionLevel = ePositionSubscriptionLevel.pslSnapshotAndUpdates;
            _cel.APIConfiguration.UseOrderSide = true;
            _cel.APIConfiguration.TimeZoneCode = eTimeZone.tzCentral; // In operator we get bar times with timeof(@) function, that always returns Line Time without DST
            _cel.APIConfiguration.ApplyDaylightSavingChanges = false;

            _cel.DataError += cqgCEL_DataError;
            _cel.CELStarted += cqgCEL_CELStarted;
            _cel.GWConnectionStatusChanged += cqgCEL_GWConnectionStatusChanged;
            _cel.InstrumentSubscribed += cqgCEL_InstrumentSubscribed;
            _cel.IncorrectSymbol += cqgCEL_IncorrectSymbol;
            _cel.OnQueryProgress += cqgCEL_OnQueryProgress;
            _cel.AccountChanged += cqgCEL_AccountChanged;
            _cel.OrderChanged += cqgCEL_OrderChanged;

            _cel.Startup();
        }

        /// <summary>
        /// Handles <see cref="CQGCEL.CELStarted"/> event.
        /// </summary>
        private void cqgCEL_CELStarted()
        {
            foreach (string symbol in _waitingSubscriptions.Keys)
            {
                requestInstrument(symbol);
            }
        }

        /// <summary>
        /// Requests CQGCEL for the specified instrument.
        /// </summary>
        /// <param name="symbol">Instrument symbol</param>
        private void requestInstrument(string symbol)
        {
            CQGInstrumentSubscriptionRequest request = _cel.CreateInstrumentRequest();
            request.Symbol = symbol;
            _cel.SubscribeNewInstrument(request);
        }

        /// <summary>
        /// Handles <see cref="CQGCEL.DataError"/> event.
        /// </summary>
        private void cqgCEL_DataError(object error, string errorDescription)
        {
            foreach (var subscribers in _waitingSubscriptions.Values)
            {
                foreach (var subscriber in subscribers)
                {
                    subscriber.OnError(errorDescription);
                }
            }

            foreach (var data in _instrumentData)
            {
                foreach (var subscriber in data.Subscribers)
                {
                    subscriber.OnError(errorDescription);
                }
            }
        }

        /// <summary>
        /// Handles <see cref="CQGCEL.InstrumentSubscribed"/> event.
        /// </summary>
        private void cqgCEL_InstrumentSubscribed(string symbol, CQGInstrument cqgInstrument)
        {
            List<IOrdersSubscriber> subscribers;
            if (_waitingSubscriptions.TryGetValue(symbol, out subscribers))
            {
                // Find or create appropriate InstrumentData
                InstrumentData instrumentData = _instrumentData.Find(x => x.InstrumentId == cqgInstrument.InstrumentID);
                if (instrumentData == null)
                {
                    instrumentData = new InstrumentData(cqgInstrument.InstrumentID);
                    _instrumentData.Add(instrumentData);

                    // If the instrument is just resolved execute order queries for it
                    foreach (CQGAccount account in _cel.Accounts)
                    {
                        instrumentData.Queries.Add(_cel.QueryOrders(account, cqgInstrument));
                    }
                }

                foreach (var subscriber in subscribers)
                {
                    instrumentData.Subscribers.Add(subscriber);
                }
                instrumentData.UpdateOrders();

                _waitingSubscriptions.Remove(symbol);
            }
        }

        /// <summary>
        /// Handles <see cref="CQGCEL.IncorrectSymbol"/> event.
        /// </summary>
        private void cqgCEL_IncorrectSymbol(string symbol)
        {
            List<IOrdersSubscriber> subscribers;
            if (_waitingSubscriptions.TryGetValue(symbol, out subscribers))
            {
                foreach (var subscriber in subscribers)
                {
                    subscriber.OnError("Symbol " + symbol + " is not recognized.");
                }

                _waitingSubscriptions.Remove(symbol);
            }
        }

        /// <summary>
        /// Handles <see cref="CQGCEL.GWConnectionStatusChanged"/> event.
        /// </summary>
        private void cqgCEL_GWConnectionStatusChanged(eConnectionStatus newStatus)
        {
            if (newStatus == eConnectionStatus.csConnectionUp)
            {
                _cel.AccountSubscriptionLevel = eAccountSubscriptionLevel.aslAccountUpdatesAndOrders;
            }
        }

        /// <summary>
        /// Handles <see cref="CQGCEL.OnQueryProgress"/> event.
        /// </summary>
        private void cqgCEL_OnQueryProgress(CQGOrdersQuery query, CQGError error)
        {
            if (query.Status == eRequestStatus.rsInProgress)
            {
                return; // ignore in progress notifications
            }

            InstrumentData data = _instrumentData.Find(x => x.Queries.Contains(query));
            if (data == null)
            {
                return;
            }

            data.Queries.Remove(query);

            if (query.Status == eRequestStatus.rsFailed)
            {
                foreach (IOrdersSubscriber subscriber in data.Subscribers)
                {
                    subscriber.OnError(error != null ? error.Description : "Order query is failed.");
                }
            }
            else if (query.Status == eRequestStatus.rsSuccess)
            {
                foreach (CQGOrder order in query.Orders)
                {
                    if (order.GWStatus == eOrderStatus.osFilled || order.GWStatus == eOrderStatus.osInOrderBook)
                    {
                        data.Orders.Add(new OrderInfo(order));
                    }
                }

                // If there is no active order queries anymore, report orders changed
                if (data.Queries.Count == 0)
                {
                    data.UpdateOrders();
                }
            }
        }

        /// <summary>
        /// Handles <see cref="CQGCEL.AccountChanged"/> event.
        /// </summary>
        private void cqgCEL_AccountChanged(eAccountChangeType changeType, CQGAccount account, CQGPosition position)
        {
            if (changeType == eAccountChangeType.actPositionsReloaded)
            {
                foreach (InstrumentData data in _instrumentData)
                {
                    // Clear account all orders
                    int ordersCount = data.Orders.Count;
                    for (int i = data.Orders.Count - 1; i >= 0; --i)
                    {
                        if (data.Orders[i].AccountId == account.GWAccountID)
                        {
                            data.Orders.RemoveAt(i);
                        }
                    }

                    if (ordersCount != data.Orders.Count)
                    {
                        data.UpdateOrders();
                    }

                    // Cancel account queries
                    foreach (CQGOrdersQuery query in data.Queries)
                    {
                        if (query.Account == account)
                        {
                            query.Cancel();
                            data.Queries.Remove(query);
                            break;
                        }
                    }

                    // Execute new queries

                    CQGInstrument instrument = _cel.Instruments[data.InstrumentId];
                    if (instrument != null)
                    {
                        data.Queries.Add(_cel.QueryOrders(account, instrument));
                    }
                }
            }
        }

        /// <summary>
        /// Handles <see cref="CQGCEL.OrderChanged"/> event.
        /// </summary>
        private void cqgCEL_OrderChanged(eChangeType changeType, CQGOrder order, CQGOrderProperties oldProperties, CQGFill cqgFill, CQGError error)
        {
            InstrumentData data = _instrumentData.Find(x => x.InstrumentId == order.Instrument.InstrumentID);
            if (data != null)
            {
                DateTime orderTime = DateTime.MaxValue;

                // For purposes of this sample ignore orders in statuses other that Filled or InOrderBook.
                bool statusOk = order.GWStatus == eOrderStatus.osFilled || order.GWStatus == eOrderStatus.osInOrderBook;

                if (!statusOk || changeType == eChangeType.ctChanged || changeType == eChangeType.ctRemoved)
                {
                    OrderInfo oldOrderInfo = data.Orders.Find(x => order.GWOrderID == x.OrderId);
                    if (oldOrderInfo != null)
                    {
                        data.Orders.Remove(oldOrderInfo);
                        orderTime = oldOrderInfo.PlaceTime;
                    }
                }

                if (statusOk)
                {
                    OrderInfo orderInfo = new OrderInfo(order);
                    data.Orders.Add(orderInfo);

                    orderTime = orderInfo.PlaceTime < orderTime ? orderInfo.PlaceTime : orderTime;
                }

                data.UpdateOrders(orderTime);
            }
        }
    }
}






