﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Text;
using CQG;

namespace SampleOrderDisplay
{
    /// <summary>
    /// Contains information required to create orders primitives
    /// </summary>
    public class OrderInfo
    {
        public string OrderId; ///< Gateway order identifier
        public int AccountId; ///< Gateway account identifier that the order belongs to
        public eOrderSide Side; ///< Order side (buy/sell)
        public eOrderType Type; ///< Order type
        public eOrderStatus Status; ///< Order status. The status could only be eOrderStatus.osFilled or eOrderStatus.osInOrderBook
        public DateTime PlaceTime; ///< Order place time
        public DateTime FillTime;  ///< Order fill time (time of the last fill)
        public double Price; ///< Order price (price of the last fill for filled orders or limit/stop price for working orders)
        public int Quantity; ///< Order quantity (filled quantity for filled orders or limit/stop quantity for working orders)
        public string Description; ///< Order description, intended to be displayed in a tooltip

        /// <summary>
        /// Initializes a new instance of <see cref="OrderInfo"/> object from the specified <see cref="CQGOrder"/>.
        /// </summary>
        public OrderInfo(CQGOrder order)
        {
            OrderId = order.GWOrderID;
            AccountId = order.Account.GWAccountID;

            Side = order.Side;
            Debug.Assert(Side != eOrderSide.osdUndefined);

            Type = order.Type;
            Status = order.GWStatus;

            PlaceTime = order.PlaceTime;

            StringBuilder sb = new StringBuilder();

            string strSide = Status == eOrderStatus.osFilled ?
                Side == eOrderSide.osdBuy ? "Bought" : "Sold" :
                Side == eOrderSide.osdBuy ? "Buy" : "Sell";

            sb.AppendFormat("{0} {1} {2}", strSide, formatOrderType(Type), formatOrderDuration(order.DurationType));

            if (Status == eOrderStatus.osFilled)
            {
                Quantity = order.FilledQuantity;
                Price = order.Fills.get_ItemByIndex(order.Fills.Count - 1).Price;
                FillTime = order.Fills.get_ItemByIndex(order.Fills.Count - 1).ServerTimestamp;

                if (order.Fills.Count > 1)
                {
                    foreach (CQGFill fill in order.Fills)
                    {
                        sb.AppendLine();
                        sb.AppendFormat(CultureInfo.InvariantCulture, " {0} at {1:.00}", fill.Quantity, fill.Price);
                    }
                }
                else
                {
                    sb.AppendFormat(CultureInfo.InvariantCulture, " {0} at {1:.00}", Quantity, Price);
                }
            }
            else
            {
                Quantity = order.Quantity - order.FilledQuantity;
                Price = Type == eOrderType.otLimit ? order.LimitPrice : order.StopPrice;
                FillTime = DateTime.MaxValue;

                sb.AppendFormat(CultureInfo.InvariantCulture, " {0} at {1:.00}", Quantity, Price);
            }

            Description = sb.ToString();
        }

        /// <summary>
        /// Formats <see cref="eOrderDuration"/> enum value to string
        /// </summary>
        private static string formatOrderDuration(eOrderDuration duration)
        {
            return duration == eOrderDuration.odDay ? "DAY" :
                duration == eOrderDuration.odATC ? "ATC" :
                duration == eOrderDuration.odATO ? "ATO" :
                duration == eOrderDuration.odFAK ? "FAK" :
                duration == eOrderDuration.odFOK ? "FOK" :
                duration == eOrderDuration.odGoodTillCanceled ? "GTC" :
                duration == eOrderDuration.odGoodTillDate ? "GTD" :
                duration == eOrderDuration.odGoodTillTime ? "GTT" : "";
        }

        /// <summary>
        /// Formats <see cref="eOrderType"/> enum value to string
        /// </summary>
        private static string formatOrderType(eOrderType orderType)
        {
            return orderType == eOrderType.otMarket ? "MKT" :
                orderType == eOrderType.otLimit ? "LMT" :
                orderType == eOrderType.otStop ? "STP" :
                orderType == eOrderType.otStopLimit ? "STPLMT" : "";
        }

    };
}
