﻿using System;
using System.Collections.Generic;
using System.Drawing;
using CQG;
using CQG.Operators;
using CQG.Operators.Primitives;

namespace SampleOrderDisplay
{
    /// <summary>
    /// Operator for Sample Order Display study. The study displays only filled and active working orders.
    /// <para>
    ///     Expression inputs:
    ///     <list type="bullet">
    ///         <item>Bar value</item>
    ///         <item>Bar time (as a count of minutes from 01.01.1900)</item>
    ///         <item>Bar duration (in seconds)</item>
    ///     </list>
    /// </para>
    /// <para>
    ///     Expression parameters:
    ///     <list type="bullet">
    ///         <item><c>buyColorFilled</c> is a color of symbol for buy filled orders (all order types)</item>
    ///         <item><c>sellColorFilled</c> is a color of symbol for sell filled orders (all order types)</item>
    ///         <item><c>buyColorWorking</c> is a color of line for buy working orders</item>
    ///         <item><c>sellColorWorking</c> is a color of line for sell working orders</item>
    ///         <item><c>filledSize</c> is relative size of filled orders symbols</item>
    ///     </list>
    /// </para>
    /// </summary>
    [CQGOperator("SampleOD(@, @, @, buyColorFilled, sellColorFilled, buyColorWorking, sellColorWorking, filledSize)")]
    public class SampleOrderDisplay : IOperator, IDisposable, IOrdersSubscriber
    {
        private CELWrapper _cqgCELWrapper;

        private IBarsRebuilder _barsRebuilder;
        private string _symbol;

        private IList<OrderInfo> _orders;

        /// <summary>
        /// Cached value of 01.01.1900 date to convert CQG bar times.
        /// </summary>
        private static DateTime _year1900 = new DateTime(1900, 1, 1);

        private Color _buyColorFilled = Color.Green; ///< Value of buyColorFilled parameter
        private Color _sellColorFilled = Color.Red;  ///< Value of sellColorFilled parameter
        private Color _buyColorWorking = Color.DarkGreen;  ///< Value of buyColorWorking parameter
        private Color _sellColorWorking = Color.DarkRed;  ///< Value of sellColorWorking parameter

        private int _filledSize = 4; ///< Value of filledSize parameter


        /// <summary>
        /// The string contains error message reported by <see cref="CELWrapper"/> or <c>null</c> if no error is occured.
        /// </summary>
        private string _error = null;

        #region IOperator implementation

        /// <summary>
        /// Implements <see cref="IOperator.Initialize"/>
        /// </summary>
        public int Initialize(IDictionary<string, string> i_params, IBarsRebuilder i_barsRebuilder)
        {
            _symbol = i_params["-RootSymbol"];

            uint colorref;
            if (uint.TryParse(i_params["buyColorFilled"], out colorref))
            {
                _buyColorFilled = colorFromCOLORREF(colorref);
            }

            if (uint.TryParse(i_params["sellColorFilled"], out colorref))
            {
                _sellColorFilled = colorFromCOLORREF(colorref);
            }

            if (uint.TryParse(i_params["buyColorWorking"], out colorref))
            {
                _buyColorWorking = colorFromCOLORREF(colorref);
            }

            if (uint.TryParse(i_params["sellColorWorking"], out colorref))
            {
                _sellColorWorking = colorFromCOLORREF(colorref);
            }
            
            int size;
            if (int.TryParse(i_params["filledSize"], out size))
            {
                _filledSize = size;
            }

            if (!String.IsNullOrEmpty(_symbol))
            {
                _cqgCELWrapper = CELWrapper.GetInstance();
                _cqgCELWrapper.SubscribeOrders(this, _symbol);
            }

            _barsRebuilder = i_barsRebuilder;

            return 2;
        }

        /// <summary>
        /// Implements <see cref="IOperator.Calculate"/>
        /// </summary>
        public void Calculate(IList<double>[] i_inputs, double[] io_outputs, ref IDeepCloneable io_state, IList<object> io_primitives)
        {
            if (_error != null)
            {
                string error = _error;
                _error = null;
                throw new ApplicationException(error);
            }

            io_outputs[0] = i_inputs[0][1];

            if (_orders != null)
            {
                DateTime timeStart = _year1900 + TimeSpan.FromMinutes(i_inputs[1][1]); // bar start time
                DateTime timeEnd = timeStart + TimeSpan.FromSeconds(i_inputs[2][1]);   // bar end time

                foreach (OrderInfo order in _orders)
                {
                    if (order.Status == eOrderStatus.osInOrderBook)
                    {
                        // For working orders we draw a horizontal line starting from the order place time to the future.
                        // We do this by adding line segment for each bar with later time, this way we ensure the line 
                        // is visible even if the order place time is out of screen and we never got request to calculate 
                        // the corresponding bar.

                        if (order.PlaceTime <= timeEnd)
                        {
                            Color color = order.Side == eOrderSide.osdBuy ? _buyColorWorking : _sellColorWorking;
                            ChartPosition position = new ChartPosition(null, order.Price);

                            if (timeStart <= order.PlaceTime) 
                            {
                                // For the first bar also write order quantity
                                TextPrimitive primText = new TextPrimitive(order.Quantity.ToString(), position)
                                {
                                    TextColor = color,
                                    VerticalAlignment = order.Side == eOrderSide.osdBuy ? VerticalAlignment.Top : VerticalAlignment.Bottom
                                };
                                io_primitives.Add(primText);
                            }

                            SymbolPrimitive primDash = new SymbolPrimitive(SymbolType.Line, position)
                            {
                                Color = color,
                                Tooltip = order.Description,
                                SizeFactor = 1
                            };
                            io_primitives.Add(primDash);
                        }
                    }
                    else if (order.FillTime >= timeStart && order.FillTime < timeEnd)
                    {
                        // For orders filled on this bar draw corresponding symbol, with type and color depending on the filled order

                        SymbolType symbolType;
                        Color color;
                        
                        if (order.Side == eOrderSide.osdUndefined)
                        {
                            symbolType = SymbolType.Cross;
                            color = Color.Gray;
                        }
                        else
                        {
                            switch (order.Type)
                            {
                                case eOrderType.otMarket:
                                    symbolType = order.Side == eOrderSide.osdBuy ? SymbolType.UpTriangle : SymbolType.DownTriangle;
                                    break;
                                case eOrderType.otLimit:
                                    symbolType = SymbolType.BigCircle;
                                    break;
                                case eOrderType.otStop:
                                case eOrderType.otStopLimit:
                                    symbolType = SymbolType.BigSquare;
                                    break;
                                default:
                                    symbolType = SymbolType.Cross;
                                    break;
                            }

                            color = order.Side == eOrderSide.osdBuy ? _buyColorFilled : _sellColorFilled;
                        }
                        
                        SymbolPrimitive prim = new SymbolPrimitive(symbolType, new ChartPosition(null, order.Price));
                        prim.Color = color;
                        prim.SizeFactor = _filledSize;

                        prim.Tooltip = order.Description;

                        io_primitives.Add(prim);
                    }
                }
            }
        }

        #endregion

        #region IOrdersSubscriber implementation

        /// <summary>
        /// Implements <see cref="IOrdersSubscriber.UpdateOrders"/> method.
        /// </summary>
        void IOrdersSubscriber.UpdateOrders(IList<OrderInfo> ordersList, DateTime? startTime)
        {
            _orders = ordersList;
            if (startTime.HasValue)
            {
                _barsRebuilder.RebuildRange(startTime.Value, DateTime.MaxValue);
            }
            else
            {
                _barsRebuilder.RebuildAll();
            }
        }

        /// <summary>
        /// Implements <see cref="IOrdersSubscriber.UpdateOrders"/> method.
        /// </summary>
        void IOrdersSubscriber.OnError(string errorMessage)
        {
            _error = errorMessage;
        }

        #endregion

        #region IDisposable implementation

        /// <summary>
        /// Implements <see cref="IDisposable.Dispose"/>
        /// </summary>
        public void Dispose()
        {
            if (_cqgCELWrapper != null)
            {
                _cqgCELWrapper.UnsubscribeOrders(this);
                _cqgCELWrapper = null;
            }
        }

        #endregion

        /// <summary>
        /// Converts COLORREF value (passed as color parameter from CQGIC) to <see cref="Color"/> value.
        /// </summary>
        private Color colorFromCOLORREF(uint color)
        {
            uint alpha = 0xFF000000 - (color & 0xFF000000);
            uint red = (color & 0xFF) << 16;
            uint green = (color & 0xFF00);
            uint blue = (color & 0xFF0000) >> 16;

            return Color.FromArgb((int)(alpha | red | green | blue));
        }
    }
}