﻿using System;
using System.Collections.Generic;
using System.Drawing;
using CQG.Operators;
using CQG.Operators.Primitives;

namespace TestOperators
{
    /// <summary>
    /// Sample operator demonstrating how to add border primitives (rectangles or ellipses) to a chart.
    /// <para>
    ///     Expression inputs:
    ///     <list type="bullet">
    ///         <item>Bar Value</item>
    ///         <item>Width (in bars count)</item>
    ///         <item>Height (in bar value units)</item>
    ///     </list>
    /// </para>
    /// <para>
    ///     Expression parameters:
    ///     <list type="bullet">
    ///         <item><c>borderType</c> - type of border (o - rectangle; 1 - ellipse)</item>
    ///     </list>
    /// </para>
    /// </summary>
    /// <remarks>
    /// The sample puts a border (either rectangle or ellipse) primitive on each bar. The width and height are specified
    /// in special inputs.
    /// </remarks>
    [CQGOperator("ShowBorder(@, @, @, borderType)")]
    public class ShowBorder : IOperator
    {
        /// <summary>
        /// Value of borderType parameter
        /// </summary>
        private BorderPrimitiveType _borderType = BorderPrimitiveType.Rectangle;

        #region IOperator implementation

        /// <summary>
        /// Implements <see cref="IOperator.Initialize"/>
        /// </summary>
        public int Initialize(IDictionary<string, string> i_params, IBarsRebuilder i_barsRebuilder)
        {
            string value = i_params["borderType"];
            if (!String.IsNullOrEmpty(value))
            {
                _borderType = int.Parse(value) > 0 ? BorderPrimitiveType.Ellipse : BorderPrimitiveType.Rectangle;
            }

            return 1;
        }

        /// <summary>
        /// Implements <see cref="IOperator.Calculate"/>
        /// </summary>
        public void Calculate(IList<double>[] i_inputs, double[] io_outputs, ref IDeepCloneable io_state, IList<object> io_primitives)
        {
            io_outputs[0] = i_inputs[0][0];

            int width = ((int)i_inputs[1][0] % 101) - 50;

            BorderPrimitive prim = new BorderPrimitive();
            prim.BorderType = _borderType;
            prim.Pen = new PrimitivePen(Color.Green);
            prim.BackgroundColor = Color.LightGreen;

            prim.Point1 = new ChartPosition();
            prim.Point2 = new ChartPosition(new ChartTime(width), i_inputs[0][0] + i_inputs[2][0]);

            prim.Tooltip = String.Format("point1 = ({0}, {1})\npoint2 = ({2}, {3})\nbackcolor = {4}\nforecolor = {5}",
                prim.Point1.Time == null ? 0 : prim.Point1.Time.BarOffset, prim.Point1.Value,
                prim.Point2.Time == null ? 0 : prim.Point2.Time.BarOffset, prim.Point2.Value, 
                prim.BackgroundColor, prim.Pen.Color);

            io_primitives.Add(prim);
        }

        #endregion
    }
}