﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Globalization;
using CQG.Operators;
using CQG.Operators.Primitives;

namespace TestOperators
{
    /// <summary>
    /// Sample operator demonstrating how to add various line primitives to a chart.
    /// <para>
    ///     Expression inputs:
    ///     <list type="bullet">
    ///         <item>Bar Value</item>
    ///         <item>Line type (0 - no line; 1 - line; 2 - ray; 3 - segment; 4 - reverse ray)</item>
    ///         <item>Horizontal shift to the second point (in bars count)</item>
    ///         <item>Vertical shift to the second point (in bar value units)</item>
    ///         <item>Color (as COLORREF value)</item>
    ///     </list>
    /// </para>
    /// <para>
    ///     Expression parameters:
    ///     <list type="bullet">
    ///         <item><c>width</c> - line width</item>
    ///     </list>
    /// </para>
    /// </summary>
    /// <remarks>
    /// The sample puts a line primitive on each bar. The first point defining the line is the bar value point.
    /// The second point, line type and line color are specified in special operator inputs.
    /// </remarks>
    [CQGOperator("ShowLine(@, @, @, @, @, width)")]
    public class ShowLines : IOperator
    {
        /// <summary>
        /// Value of width parameter
        /// </summary>
        private double _width = 1.0;
        
        #region IOperator implementation

        /// <summary>
        /// Implements <see cref="IOperator.Initialize"/>
        /// </summary>
        public int Initialize(IDictionary<string, string> i_params, IBarsRebuilder i_barsRebuilder)
        {
            double width;
            if (double.TryParse(i_params["width"], NumberStyles.Float, CultureInfo.InvariantCulture, out width))
            {
                _width = width;
            }

            return 1;
        }

        /// <summary>
        /// Implements <see cref="IOperator.Calculate"/>
        /// </summary>
        public void Calculate(IList<double>[] i_inputs, double[] io_outputs, ref IDeepCloneable io_state, IList<object> io_primitives)
        {
            io_outputs[0] = i_inputs[0][0];

            LinePrimitive prim = new LinePrimitive();

            // Set line's points
            int dx = (int)i_inputs[2][0];
            double y = i_inputs[0][0] + i_inputs[3][0];

            prim.Point1 = new ChartPosition();
            prim.Point2 = new ChartPosition(new ChartTime(dx), y);

            int lineType = (int)i_inputs[1][0] % 5;
            if (lineType == 0)
            {
                return;
            }

            if (lineType == 2 || lineType == 3)
            {
                prim.EndsInPoint1 = true;
            }

            if (lineType == 3 || lineType == 4)
            {
                prim.EndsInPoint2 = true;
            }

            // Set line pen
            int penStyle = (int)i_inputs[1][0] % 6; // PS_SOLID = 0; PS_DASH = 1; PS_DOT = 2; PS_DASHDOT = 3
            if (penStyle >= 4)
            {
                penStyle = 5; // PS_NULL
            }

            Color color = colorFromCOLORREF((uint)i_inputs[4][0]);

            prim.Pen = new PrimitivePen(color, penStyle, _width);

            // Set the primitive's tooltip
            prim.Tooltip = String.Format("Line type = {0}\nColor = {1}",
                lineType == 1 ? "Line" :
                lineType == 2 ? "Ray" :
                lineType == 3 ? "Segment" :
                lineType == 4 ? "ReverseRay" : "Unknown",
                color);

            io_primitives.Add(prim);
        }

        #endregion

        /// <summary>
        /// Convert COLORREF value passed from CQGIC to <see cref="Color"/>
        /// </summary>
        private Color colorFromCOLORREF(uint color)
        {
            uint alpha = 0xFF000000 - (color & 0xFF000000);
            uint red = (color & 0xFF) << 16;
            uint green = (color & 0xFF00);
            uint blue = (color & 0xFF0000) >> 16;

            return Color.FromArgb((int)(alpha | red | green | blue));
        }
    }
}