﻿using System;
using System.Collections.Generic;
using System.Drawing;
using CQG.Operators;
using CQG.Operators.Primitives;

namespace TestOperators
{
    /// <summary>
    /// Sample operator demonstrating how to add symbol primitives to a chart.
    /// <para>
    ///     Expression inputs:
    ///     <list type="bullet">
    ///         <item>Bar Value</item>
    ///         <item><see cref="SymbolType"/> to use for the primitive</item>
    ///         <item>Symbol size factor</item>
    ///         <item>Symbol color (as COLORREF value)</item>
    ///     </list>
    /// </para>
    /// </summary>
    /// <remarks>
    /// The sample puts a symbol on each bar. Symbol type, symbol size and symbol color are specified
    /// in special operator inputs.
    /// </remarks>    
    [CQGOperator("ShowSymbol(@, @, @, @)")]
    public class ShowSymbol : IOperator
    {
        private int[] _symbolTypes = (int[])Enum.GetValues(typeof(SymbolType));

        #region IOperator implementation

        /// <summary>
        /// Implements <see cref="IOperator.Initialize"/>
        /// </summary>
        public int Initialize(IDictionary<string, string> i_params, IBarsRebuilder i_barsRebuilder)
        {
            return 1;
        }

        /// <summary>
        /// Implements <see cref="IOperator.Calculate"/>
        /// </summary>
        public void Calculate(IList<double>[] i_inputs, double[] io_outputs, ref IDeepCloneable io_state, IList<object> io_primitives)
        {
            io_outputs[0] = i_inputs[0][0];

            SymbolType symbolType = (SymbolType)_symbolTypes[(((int)i_inputs[1][0]) % _symbolTypes.Length)];

            SymbolPrimitive prim = new SymbolPrimitive(symbolType, new ChartPosition());
            prim.SizeFactor = Math.Min((int)i_inputs[2][0], 100);

            prim.Color = colorFromCOLORREF((uint)i_inputs[3][0]);

            prim.Tooltip = symbolType.ToString();
            if (symbolType == SymbolType.BigDiamond)
            {
                prim.Tooltip += Environment.NewLine + " size = " + prim.SizeFactor; // tooltip with 2 lines
            }
            else if (symbolType == SymbolType.BigSquare)
            {
                prim.Tooltip += "\nsize = " + prim.SizeFactor; // tooltip with 2 lines
                prim.Color = Color.Red;
            }
            else if (symbolType == SymbolType.BigCircle)
            {
                prim.Color = Color.FromArgb(20, 10, 255, 20);
            }
            else if (symbolType == SymbolType.UpBigArrow)
            {
                prim.Tooltip += "\nsize = " + prim.SizeFactor + "\ncolor = " + prim.Color.ToString(); // tooltip with 3 lines
            }

            io_primitives.Add(prim);
        }

        #endregion

        /// <summary>
        /// Convert COLORREF value passed from CQGIC to <see cref="Color"/>
        /// </summary>
        private Color colorFromCOLORREF(uint color)
        {
            uint alpha = 0xFF000000 - (color & 0xFF000000);
            uint red = (color & 0xFF) << 16;
            uint green = (color & 0xFF00);
            uint blue = (color & 0xFF0000) >> 16;

            return Color.FromArgb((int)(alpha | red | green | blue));
        }
    }
}