﻿using System;
using System.Collections.Generic;
using System.Drawing;
using CQG.Operators;
using CQG.Operators.Primitives;

namespace TestOperators
{
    /// <summary>
    /// Sample operator demonstrating how to add text primitives to a chart.
    /// <para>
    ///     Expression inputs:
    ///     <list type="bullet">
    ///         <item>Bar Value</item>
    ///         <item>Index of text sample: there are 4 samples with different background, size etc.</item>
    ///         <item>Indicates whether the primitive should be added: 1 - yes, 0 - no</item>
    ///     </list>
    /// </para>
    /// <para>
    ///     Expression parameters:
    ///     <list type="bullet">
    ///         <item><c>backOpacity</c> - opacity for text background</item>
    ///         <item><c>foreOpacity</c> - opacity for text foreground</item>
    ///     </list>
    /// </para>    
    /// </summary>
    [CQGOperator("ShowText(@, @, @, backOpacity, foreOpacity)")]
    public class ShowText : IOperator
    {
        private int _foreOpacity = 0xFF; ///< Value of foreOpacity parameter
        private int _backOpacity = 0xFF; ///< Value of backOpacity parameter

        #region IOperator implementation

        /// <summary>
        /// Implements <see cref="IOperator.Initialize"/>
        /// </summary>
        public int Initialize(IDictionary<string, string> i_params, IBarsRebuilder i_barsRebuilder)
        {
            int value;
            if (int.TryParse(i_params["foreOpacity"], out value))
            {
                _foreOpacity = value & 0xFF;
            }

            if (int.TryParse(i_params["backOpacity"], out value))
            {
                _backOpacity = value & 0xFF;
            }

            return 1;
        }

        /// <summary>
        /// Implements <see cref="IOperator.Calculate"/>
        /// </summary>
        public void Calculate(IList<double>[] i_inputs, double[] io_outputs, ref IDeepCloneable io_state, IList<object> io_primitives)
        {
            io_outputs[0] = i_inputs[0][0];

            if (i_inputs[2][0] == 0.0)
            { 
                return;
            }

            TextPrimitive prim = new TextPrimitive(io_outputs[0].ToString());

            int index = (int)i_inputs[1][0] % 4;
            switch (index)
            {
                case 0: // text with a border and background
                    prim.BorderColor = Color.FromArgb(_foreOpacity, Color.Red);
                    prim.BackgroundColor = Color.FromArgb(_backOpacity, Color.LightBlue);
                    prim.TextColor = Color.FromArgb(_foreOpacity, Color.Blue);
                    break;
                case 1: // text with thick border and padding, it's placed on the previous bar
                    prim.Point = new ChartPosition(new ChartTime(-1));

                    prim.BorderColor = Color.FromArgb(_foreOpacity, Color.LightBlue);
                    prim.BorderWidth = 3;
                    prim.BackgroundColor = Color.FromArgb(_backOpacity, Color.Gray);
                    prim.TextColor = Color.FromArgb(_foreOpacity, Color.White);
                    prim.Padding = 3.1;
                    break;
                case 2: // text with size factor 2
                    prim.TextColor = Color.FromArgb(_foreOpacity, Color.Red);
                    prim.SizeFactorMultiplier = 2.0;
                    break;
                case 3: // text with size factor 3, it has border but uses transparent background
                    prim.BorderColor = Color.FromArgb(_foreOpacity, Color.Purple);
                    prim.BorderWidth = 2;
                    prim.TextColor = Color.FromArgb(_foreOpacity, Color.PapayaWhip);
                    prim.SizeFactorMultiplier = 3.0;
                    break;
            }

            io_primitives.Add(prim);
        }

        #endregion
    }
}
