// CurrentSecond.cpp is part of CQG COM Operators SDK Sample

#include "stdafx.h"
#include <limits>

#include "CurrentSecond.h"
#include "comstr.h"

#define TIMER_EVENT 1

// CCurrentSecond
CCurrentSecond::CCurrentSecond()
   : m_updatePeriodSeconds(INVALID_INTEGER)
   , m_timerQueue(NULL)
   , m_timer(NULL)
{
}

CCurrentSecond::~CCurrentSecond()
{
   if (m_timerQueue)
   {
      ATLVERIFY(DeleteTimerQueue(m_timerQueue));
   }
}

STDMETHODIMP CCurrentSecond::GetTitle(LPWSTR * o_title, LPWSTR * o_error)
{
   wchar_t s[128] = L"";

   // NOTE amount of %%s qualifiers shall correspond to GetInputCount call result
   _snwprintf(s, sizeof(s)/sizeof(s[0]), L"CurrentSecond(%%s, %d)", m_updatePeriodSeconds);

   return CreateReturnString(o_title, s);
}

STDMETHODIMP CCurrentSecond::Initialize(short i_ii, LPWSTR i_expression, LPWSTR *)
{
   // i_ii = 0 - input value (variable)
   // i_ii = 1 - update period, seconds

   switch (i_ii)
   {
      case 1: // i_ii = 1 - CQGIC asks to initialize update period, seconds
         m_updatePeriodSeconds = ConstantInt(i_expression);
         if (m_updatePeriodSeconds <= 0 || m_updatePeriodSeconds > 30) // intraday only
         {
            m_updatePeriodSeconds = INVALID_INTEGER;
            return E_INVALIDARG;
         }
         break;

      default: // not required to handle other initializers
         break;

   }

   return S_OK;
}

STDMETHODIMP CCurrentSecond::GetInputCount(short * o_inputCount, LPWSTR * o_error)
{
   if (o_inputCount)
   {
      *o_inputCount = 1; // fake input
      return S_OK;
   }

   return E_INVALIDARG;
}

STDMETHODIMP CCurrentSecond::GetOutputCount(short * o_outputCount, LPWSTR * o_error)
{
   if (o_outputCount)
   {
      *o_outputCount = 1; // second in minute 
      return S_OK;
   }

   return E_INVALIDARG;
}

STDMETHODIMP CCurrentSecond::GetOutputBarType(short i_ii, LPWSTR * o_barTypeText, LPWSTR * o_error)
{
   if (i_ii == 0)
   {
      return CreateReturnString(o_barTypeText, L"curve1");
   }

   return S_FALSE;
}

STDMETHODIMP CCurrentSecond::GetParameterCount(short * o_parmCount, LPWSTR * o_error)
{
   if (o_parmCount)
   {
      // GetParameterCount is supposed to return amount of constant operands.
      *o_parmCount = 1;
      return S_OK;
   }

   return E_INVALIDARG;
}

STDMETHODIMP CCurrentSecond::GetParameter(short i_ii, LPWSTR * o_parameter, LPWSTR * o_error)
{
   wchar_t s[128] = L"";

   // the i_ii passed in here is relative to zero. we shall get here i_ii from range [0; GetParameterCount() - 1]
   switch (i_ii)
   {
      case 0: // i_ii = 0 - CQGIC asks us to reply value of m_frameMinutes
         _snwprintf(s, sizeof(s) / sizeof(s[0]), L"%d", m_updatePeriodSeconds);
         break;

      default: // CQGIC asks about something we don't know
         return E_INVALIDARG;
         break;

   }

   return CreateReturnString(o_parameter, s);
}

STDMETHODIMP CCurrentSecond::GetExtraStepCount(long * o_extraStepCount, LPWSTR * o_error)
{
   if (o_extraStepCount)
   {
      *o_extraStepCount = 0;
      return S_OK;
   }

   return E_INVALIDARG;
}

STDMETHODIMP CCurrentSecond::InitializeStep(int IsCalcLatest, int i_updateStepIndex, int i_sizeIn, double * i_lastValidInput)
{
   UNREFERENCED_PARAMETER(IsCalcLatest);
   UNREFERENCED_PARAMETER(i_updateStepIndex);

   if (i_sizeIn < 2)
   {
      return E_INVALIDARG;
   }

   return S_OK;
}

namespace
{
   long getSecondOfDay() // returns second of current day
   {
      SYSTEMTIME time = { 0 };
      GetSystemTime(&time);

      return time.wHour * 60 * 60 + time.wMinute * 60 + time.wSecond;
   }
}

STDMETHODIMP CCurrentSecond::Calculate(int i_stepIndex, int i_sizeIn, double * i_input, int i_sizeOut, double * o_output)
{
   if (i_sizeIn < 1 || i_sizeOut < 1 || m_updatePeriodSeconds <= 0)
   {
      return E_INVALIDARG;
   }

   o_output[0] = getSecondOfDay();

   return S_OK;
}

STDMETHODIMP CCurrentSecond::GetOutputScaleType(short i_ii, short * o_scaleType, short * o_index1, short * o_index2, LPWSTR *)
{
   if (o_scaleType && o_index1 && o_index2)
   {
      *o_scaleType = OSD_NEW;
      *o_index1 = 0;
      *o_index2 = 0;

      return S_OK;
   }

   return E_INVALIDARG;
}

STDMETHODIMP CCurrentSecond::InitializeConstants(int i_sizeIn, double * i_input)
{
   // CQGIC gives a chance to initialize following constants here:
   // i_input[0] - minimum value change (e.g. tick size for price scales)
   // i_input[1] - price to dollar multiplier (for price scales)

   return S_OK;
}

namespace
{
   VOID CALLBACK TimerCallback(PVOID i_lpParam, BOOLEAN i_timerOrWaitFired)
   {
      ATLASSERT(i_lpParam);

      if (i_lpParam && i_timerOrWaitFired)
      {
         ICQGOpUpdateSink *sink = reinterpret_cast<ICQGOpUpdateSink*>(i_lpParam);
         sink->RebuildAll();
      }
   }
}

STDMETHODIMP CCurrentSecond::RegisterUpdateSink(ICQGOpUpdateSink *i_sink)
{
   if (i_sink)
   {
      if (i_sink != m_updateSink)
      {
         if (m_timerQueue)
         {
            ATLVERIFY(DeleteTimerQueue(m_timerQueue));
            m_timerQueue = NULL;
            m_timer = NULL;
         }

         m_updateSink = i_sink;
         m_timerQueue = CreateTimerQueue();

         if (m_timerQueue)
         {
            ATLVERIFY(CreateTimerQueueTimer(&m_timer, m_timerQueue, (WAITORTIMERCALLBACK)TimerCallback,
               m_updateSink.p, 0 /* call first time immediately */, 1000 * m_updatePeriodSeconds, WT_EXECUTEDEFAULT));
         }
      }

      return S_OK;
   }

   return E_INVALIDARG;
}

STDMETHODIMP CCurrentSecond::UnregisterUpdateSink(ICQGOpUpdateSink *i_sink)
{
   if (i_sink == m_updateSink.p && m_updateSink.p)
   {
      if (m_timerQueue)
      {
         ATLVERIFY(DeleteTimerQueue(m_timerQueue));
         m_timerQueue = NULL;
         m_timer = NULL;
      }

      m_updateSink.Release();
      return S_OK;
   }

   return E_INVALIDARG;
}
