// MyMinMaxRange.cpp is part of CQG COM Operators SDK Sample

#include "stdafx.h"
#include <limits>

#include "MyMinMaxRange.h"
#include "comstr.h"

CMyMinMaxRange::CMyMinMaxRange()
{
   // reference the COM object of nested operator with the known CLSID.
   m_minmaxOp.CoCreateInstance(CLSID_MyMinMax);
}

STDMETHODIMP CMyMinMaxRange::GetTitle(LPWSTR * o_title, LPWSTR * o_error)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   CComHeapPtr<wchar_t> parm1;
   HRESULT hr = m_minmaxOp->GetParameter(0, &parm1, o_error);

   if (SUCCEEDED(hr))
   {
      // NOTE amount of %%s specifiers shall correspond to GetInputCount call result

      wchar_t s[128] = { 0 };
      _snwprintf(s, _countof(s) - 1, L"MyMinMaxRange(%%s, %S)", parm1.m_pData);

      return CreateReturnString(o_title, s);
   }

   return hr;
}

STDMETHODIMP CMyMinMaxRange::Initialize(short i_ii, LPWSTR i_expression, LPWSTR * o_error)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // method Initialize is to be called for each parameter and each input expression.

   // if our operator reports 1 input, then one and only input is to be initialized with call to
   // Initialize with i_ii equal to 0. if our operator reports 10 inputs, then these 10 inputs are to be
   // initialized with 10 calls to Initialize with i_ii: [0...9]

   // after N inputs initilized, all operator parameters which were typed in CQGIC Formula Builder
   // are to be initilized with calls to Initialize with i_ii starting with value of N.

   // in the code below, we only do init first parameter 'period' with index i_ii == 1 because we have one
   // input and don't want specific initialization code for it (for i_ii == 0).
   // we don't store our 'period' value and delegate this parameter to the nested MyMinMax operator.

   if (i_ii == 1) 
   {
      return m_minmaxOp->Initialize(1, i_expression, o_error);
   }

   return S_OK;
}

STDMETHODIMP CMyMinMaxRange::GetInputCount(short * o_inputCount, LPWSTR * o_error)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // for simplicity, just return 1 here (we known MyMinMax internals well enough), but in general case
   // it would be better to ask referenced operator as well. if ask referenced operator,
   // GetTitle should be adapted for this as well.

   if (o_inputCount)
   {
      *o_inputCount = 1;
      return S_OK;
   }

   return E_INVALIDARG;
}

STDMETHODIMP CMyMinMaxRange::GetOutputCount(short * o_outputCount, LPWSTR * o_error)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   if (o_outputCount)
   {
      // we output only 1 value (our operator calculates the difference of nested operator MyMinMax outputs).
      // notice that nested operator itself has 2 outputs. this fact is assert'ed during Calculate work.
      *o_outputCount = 1;
      return S_OK;
   }

   return E_INVALIDARG;
}

STDMETHODIMP CMyMinMaxRange::GetOutputBarType(short i_ii, LPWSTR * o_barTypeText, LPWSTR * o_error)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // GetOutputBarType is to be called for each output in amount declared by GetOutputCount.
   // first output's bar type is asked by call to GetOutputBarType with index i_ii == 0 and so on.

   // general practice is to return 'curveN' text value for Nth output.
   // where is an option to return bar types like open/high/low/close, but it makes sense
   // only for operators which try to compose OHLC-like bar outputs (and so would reply more
   // than 1 from GetOutputCount) - that can be useful in advanced charting related to QFormulas or
   // to use same operator with output selector operator in Formula Builder (like "MyMinMax (@,10) : curve2")

   // please note that CQG IC expects limited set of literals here. strings of free form will
   // be most probably considered as same output, so better use curve1 - curve55 semantics or
   // open/high/low/close (see above), if you don't exactly know what you do.

   if (i_ii == 0)
   {
      return CreateReturnString(o_barTypeText, L"curve1");
   }

   return S_FALSE;
}

STDMETHODIMP CMyMinMaxRange::GetParameterCount(short * o_parmCount, LPWSTR * o_error)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // sometimes CQG IC wants to know all the parameters of the COM operator, just to be able
   // to compare them to understand whether two operators are the same or can produce differing
   // output. in such case, CQG IC asks operator the amount of important parameters it has through
   // method GetParameterCount and then asks to reply each one's string rep through GetParameter.

   // our MyMinMaxRange fully depends on MyMinMax operator and so delegates these questions there.

   if (o_parmCount)
   {
      return m_minmaxOp->GetParameterCount(o_parmCount, o_error);
   }

   return E_INVALIDARG;
}

STDMETHODIMP CMyMinMaxRange::GetParameter(short i_ii, LPWSTR * o_parameter, LPWSTR * o_error)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // see comments in GetParameterCount.

   return m_minmaxOp->GetParameter(i_ii, o_parameter, o_error);
}

STDMETHODIMP CMyMinMaxRange::GetExtraStepCount(long * o_extraStepCount, LPWSTR * o_error)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // GetExtraStepCount is important for CQG IC to understand "how much bars would this operator
   // require to start produce correct value?" for example, Simple Moving Average with period 10
   // requires current and 9 bars from past, so it returns 10 (just period value). some studies
   // like Exponential Moving Average require some amount of bars into past to provide some error
   // which wouldn't exceed the specified level.

   // in our case we just forward this question to the nested operator.

   return m_minmaxOp->GetExtraStepCount(o_extraStepCount, o_error);
}

STDMETHODIMP CMyMinMaxRange::InitializeStep(int i_sizeIn, double * i_lastValidInput)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // when CQG IC is going to run a brand new cycle of calculations ('stepping'), it asks operator to InitializeStep.
   // this is time to clear State objects and to be ready to start processing of the data, which will include
   // GetExtraStepCount of bars from past. futher Calculate calls will start from most old bar till most recent. 

   return m_minmaxOp->InitializeStep(i_sizeIn, i_lastValidInput);
}

STDMETHODIMP CMyMinMaxRange::Calculate(int i_sizeIn, double * i_input, int i_sizeOut, double * o_output)
{
   o_output[0] = INVALID_VALUE;

   if (!isOK())
   {
      return E_FAIL;
   }

   // at each bar Calculation, CQG IC calls COM operator's Calculate method to update its state (if any)
   // and return outputs.

   // size of i_input array is i_sizeIn and shall correspond to GetInputCount.
   // size of i_output array is i_sizeOut and shall correspond to GetOutputCount.

   // as it was already mentioned in InitializeStep method comments, CQG IC gives guarantee that the calculation
   // cycle will be started in consistency with method GetExtraStepCount result. so, until your operator is not
   // ready to return stable(***) values, you can collect data or calculate some 'startup' value and store
   // it in your State object.

   // when you understand that you have enough data to yield stable output, you should output some value into
   // o_output array. during startup process, output should be filled with INVALID_VALUE.

   // *** stable output means that operator output should not depend on calculation starting point and
   //     this is operator's only responsibility to implement algorithm and GetExtraStepCount correctly,
   //     otherwise, your study values at chart have all chances to change in past unexpectedly.

   if (i_sizeIn < 1 || i_sizeOut < 1)
   {
      return E_INVALIDARG;
   }

   CComHeapPtr<wchar_t> error;
   short nestedOutputCount = 0;

   // let's assert that our nested operator satisfy its semantics, i.e., it returns high/low levels - 2 outputs.

   HRESULT hr = m_minmaxOp->GetOutputCount(&nestedOutputCount, &error);

   if (!SUCCEEDED(hr) || 2 != nestedOutputCount)
   {
      return SUCCEEDED(hr) ? E_FAIL : hr;
   }

   // calculate max and min levels using nested operator.

   double minmax[2] = { 0 };
   hr = m_minmaxOp->Calculate(i_sizeIn, i_input, 2, minmax);

   if (!SUCCEEDED(hr))
   {
      return hr;
   }

   // reply the levels range if both leves are defined.
   if (minmax[0] != INVALID_VALUE && minmax[1] != INVALID_VALUE)
   {
      o_output[0] = minmax[1] - minmax[0];
   }

   return S_OK;
}

STDMETHODIMP CMyMinMaxRange::SaveState(short i_index)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // CQG IC is free to ask COM operator to stop calculations at any moment and store indexed state.
   // the index for the state is 'i_index'. in most cases, CQG IC would ask to do this for i_index == 0 only.

   // when SaveState is called, COM operator shall copy all the state data into internal storage.

   // after state with i_index = x is saved successfully, CQG IC is free to ask to revert operator by call to
   // RestoreState(x), and your operator must assign current state (the one which is used in Calculate) to
   // state(x) from operator's internal storage.

   // MyMinMaxRange itself has no state, so just forward State management to MyMinMax nested operator.

   return m_minmaxOp->SaveState(i_index);
}

STDMETHODIMP CMyMinMaxRange::RestoreState(short i_index)
{
   if (!isOK())
   {
      return E_FAIL;
   }

   // see comments in SaveState method.

   return m_minmaxOp->RestoreState(i_index);
}

