// MyRangeMarker.h is part of CQG COM Operators SDK Sample

#pragma once

#include <vector>
#include <deque>
#include <utility>

#include "resource.h"       // main symbols

#include "RangeMarkerSample.h"
#include "MyOperatorBase.h"

struct Range
{
   double hi;     // high level of the range
   double lo;     // low level of the range

   int timeStart;        // time of range start
   int timeEndExclusive; // theoretical end of range
   int timeEndInclusive; // fact time of range end, displayed

   Range(double i_hi = INVALID_VALUE, double i_lo = INVALID_VALUE, int i_timeStart = INVALID_INTEGER, int i_timeEnd = INVALID_INTEGER)
      : hi(i_hi), lo(i_lo), timeStart(i_timeStart), timeEndExclusive(i_timeEnd), timeEndInclusive(i_timeStart)
      { }

   bool IsValid() const
      { return INVALID_INTEGER != timeStart && INVALID_INTEGER != timeEndExclusive &&
               INVALID_VALUE != hi && INVALID_VALUE != lo &&
               hi >= lo && timeStart <= timeEndExclusive; }

   bool Update(double i_newValue, int i_newTimeMinutes)
   {
      ATLASSERT(IsValid());

      bool res = false;

      if (INVALID_VALUE == hi || i_newValue > hi)
      {
         hi = i_newValue;
         res = true;
      }

      if (INVALID_VALUE == lo || i_newValue < lo)
      {
         lo = i_newValue;
         res = true;
      }

      if (i_newTimeMinutes > timeEndInclusive)
      {
         timeEndInclusive = i_newTimeMinutes;
         res = true;
      }

      return res;
   }

   // this method replies true if the range we cover has intersection with the range passed in. supposes i_startTime <= i_endTime.
   bool HasIntersection(int i_startTime, int i_endTime) const
      { return IsValid() && i_endTime >= timeStart && i_startTime <= timeEndExclusive; }

   // this method updates primitive structures vector with this Range element in accordance with flags passed in.
   // it replies true if it pushed at least 1 primitive into the vector passed in.
   bool AppendPrimitives(std::vector<Primitive>& io_primitives,
      bool i_shading, COLORREF i_colorShading, bool i_label, bool i_norm, double i_tick)  const;

};

struct CMyRangeMarkerState
{
   std::pair<double, int> lastValidInput;

   std::deque<std::pair<double, int> > timeAndValueBuffer; // this queue is used for algorithm startup only
   std::vector<Range> ranges;

   // this method updates the state with new time and value in accordance with frame specified. returns true if some range is changed.
   bool ProcessInput(double i_newValue, int i_newTimeMinutes, int i_frameMinutes, size_t i_maxBufSize);

};


// CMyRangeMarker
class ATL_NO_VTABLE CMyRangeMarker
   : public CComObjectRootEx<CComSingleThreadModel>
   , public CComCoClass<CMyRangeMarker, &CLSID_MyRangeMarker>
   , public CMyOperatorBase<CMyRangeMarkerState>
   , public ICQGOp2
   , public ICQGPrimitive2
{
   typedef CMyOperatorBase<CMyRangeMarkerState> MyBase;

public:
   CMyRangeMarker();

   DECLARE_REGISTRY_RESOURCEID(IDR_MYRANGEMARKER)

   DECLARE_NOT_AGGREGATABLE(CMyRangeMarker)

   BEGIN_COM_MAP(CMyRangeMarker)
       COM_INTERFACE_ENTRY(ICQGOp)
       COM_INTERFACE_ENTRY(ICQGOp2)
       COM_INTERFACE_ENTRY(ICQGPrimitive2)
   END_COM_MAP()

   DECLARE_PROTECT_FINAL_CONSTRUCT()

   HRESULT FinalConstruct()
      { return S_OK; }

   void FinalRelease()
      { }

public:

   ///@name ICQGOp interface methods
   ///@{
   STDMETHOD(GetTitle)(LPWSTR * o_title, LPWSTR * o_error);
   STDMETHOD(Initialize)(short i_ii, LPWSTR i_expression, LPWSTR * o_error);
   STDMETHOD(GetInputCount)(short * o_inputCount, LPWSTR * o_error);
   STDMETHOD(GetOutputCount)(short * o_outputCount, LPWSTR * o_error);
   STDMETHOD(GetOutputBarType)(short i_ii, LPWSTR * o_barTypeText, LPWSTR * o_error);
   STDMETHOD(GetParameterCount)(short * o_parmCount, LPWSTR * o_error);
   STDMETHOD(GetParameter)(short i_ii, LPWSTR * o_parameter, LPWSTR * o_error);
   STDMETHOD(GetExtraStepCount)(long * o_extraStepCount, LPWSTR * o_error);
   ///@}

   ///@name ICQGOp2 interface methods
   ///@{
   
   STDMETHOD(InitializeStep)(int IsCalcLatest, int i_updateStepIndex, int i_sizeIn, double * i_lastValidInput);
   STDMETHOD(Calculate)(int i_stepIndex, int i_sizeIn, double * i_input, int i_sizeOut, double * o_output);
   STDMETHOD(InitializeConstants)(int i_sizeIn, double * i_input);

   STDMETHOD(NeedsInputsFilledIn)(int * o_returnFlag)
      { *o_returnFlag = TRUE; return S_OK; }
   STDMETHOD(CanHaveGapsInOutputs)(int * o_returnFlag)
      { *o_returnFlag = TRUE; return S_OK; }
   STDMETHOD(IsDependentOnStartTime)(int * o_returnFlag)
      { *o_returnFlag = TRUE; return S_OK; }
   STDMETHOD(ScrollValueSet)(int * o_returnFlag)
      { *o_returnFlag = 3; return S_OK; }

   ///@}

   ///@name ICQGPrimitive2 interface methods
   ///@{
   STDMETHOD(HavePrimitivesChanged)(short * o_changed);
   STDMETHOD(GetPrimitives)(ChartTime i_startTime, ChartTime i_endTime,
      int * o_count, Primitive * * o_primitives);
   ///@}

private:

   // this is special constant which is initialized by CQGIC with call to Initialize with -1 index. we catch it to have
   // ability to return this value from GetExtraStepCount as soon as we don't know how many bars are required to fit
   // m_frameMinutes of bars absolute time. this value can be altered by user through System Preferences -> Limits tab.
   //
   // IMPORTANT NOTE ON PERFORMANCE. if your operator knows the period of calculation it must be returned here, not
   // max extra steps like here for this specific sample operator. For example, simple moving average algorithm with
   // period = N always knows that it is enough N bars to have to start yielding valid values. So return N here.
   // Another (trivial) example - you for example reinventing "operator + (x, y)" - and so you have to return 1 here,
   // because your output depends only on exactly ONE bar, nothing from past.

   int m_cqgicMaxExtraSteps;


   int m_frameMinutes; // time frame size in minutes to be shaded/markered in its min/max range

   bool m_enableShading;    // true to enable shading primitives output
   COLORREF m_colorShading; // RGB color of shading

   bool m_enableLabel;          // true to enable text labels
   bool m_enableLabelNormalize; // true to enable text labels size normalization

   // stores information whether primitives array was changed during last calculation cycle
   bool m_primitivesChanged;

   double m_tickSize; // min val change, see comments in InitializeConstants


};

OBJECT_ENTRY_AUTO(__uuidof(MyRangeMarker), CMyRangeMarker)
